#include <cmath>
#include <algorithm>
#include "barycentric.h"
#include "basic.h"
#include "bezier.hpp"
#include "global.hpp"


///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Compute Barycentric Basis Functions for each vertex of the polygon Poly
CMatrix* BaseFunctions( const vector<CPoint>& sample, const CPoly& Poly, unsigned int nSubd )
{
	////////////////////////////////////////////////////////////////
	// construct QuadTree
	vector < unsigned int > plines( Poly.size() );
	for ( unsigned int i = 0; i < Poly.size(); i++ )
		plines[i] = i;

	QUADTREE::Quadtree quadTree( &Poly, plines );
	////////////////////////////////////////////////////////////////

	CMatrix* pMatrix= new CMatrix( sample.size(), Poly.size() );

	for ( unsigned int i = 0; i < sample.size(); i++ )
	{
		if ( i % 4 == 0 || i == sample.size() - 1 )
			cout << "progressing " << i + 1 << "/" << sample.size() << "\r";

		CVector* weight = Barycentric( quadTree, sample[i], Poly );
		pMatrix->setRow( i, *weight );
	}
	cout << endl;

	pMatrix->Transpose(); // now pMatrix is a Length[Poly] by Length[Sample] matrix

	return pMatrix;
}

//'Barycentric' compute the Barycentric coordinate for a point on the boundary or in the interior of an arbitrary polygon
CVector* Barycentric( QUADTREE::Quadtree& quadTree, const CPoint& X, const CPoly& Poly )
{
	CVector* weight = new CVector( Poly.size() );
/*
	bool onPoly = OnPolygon( X, Poly, *weight );

	if ( onPoly )
		return weight;

	if ( !InsidePolygon( X, Poly ) )
		return weight;
*/
	for ( unsigned int i = 0; i < Poly.size(); i++ )
		IntegrateAllEdges( quadTree, i, X, Poly, *weight );

	// normalize weight
	weight->normalize();

	return weight;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////



///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Compute Barycentric Basis Functions for each vertex of the polygon Poly
CMatrix* approBaseFunctions( const vector<CPoint>& sample, const CPoly& Poly, unsigned int nSubd )
{
	////////////////////////////////////////////////////////////////
	// quadratic b-spline curve
	/*
	CMatrix* subdMatrix = Bspline( *Poly2Matrix( Poly ), nSubd );
	CPoly* subdPoly = Matrix2Poly( *subdMatrix );

	CMatrix* bsplineWeight = BsplineWeight( Poly.size(), nSubd );
	*/
	////////////////////////////////////////////////////////////////
	// adaptive quadratic bezier curve
	AdaptiveBezier *adpBezier = new AdaptiveBezier( Poly.size() );
	adpBezier->bezier( *Poly2Matrix( Poly ), nSubd );

	CPoly* subdPoly = Matrix2Poly( *adpBezier->subdPoly );

	CMatrix* bsplineWeight = adpBezier->subdWeight;


	////////////////////////////////////////////////////////////////
	// construct QuadTree
	vector < unsigned int > plines( subdPoly->size() );
	for ( unsigned int i = 0; i < subdPoly->size(); i++ )
		plines[i] = i;

	QUADTREE::Quadtree quadTree( subdPoly, plines );
	////////////////////////////////////////////////////////////////
	// for test only below
	//CVector* weight = approBarycentric( quadTree, CPoint(0.12, 1.3), Poly.size(), *subdPoly, *bsplineWeight );
	//cout << *weight << endl;
	//exit(0);
	////////////////////////////////////////////////////////////////

	CMatrix* pMatrix= new CMatrix( sample.size(), Poly.size() );

	//#pragma omp parallel for
	for ( int i = 0; i < sample.size(); i++ )
	{
		if ( (i & 15) == 0 || i == sample.size() - 1 )
			cout << "progressing " << i + 1 << "/" << sample.size() << "\r";

		CVector* weight = approBarycentric( quadTree, sample[i], Poly.size(), *subdPoly, *bsplineWeight );
		pMatrix->setRow( i, *weight );
	}
	cout << endl;

	pMatrix->Transpose(); // now pMatrix is a Length[Poly] by Length[Sample] matrix

	return pMatrix;
}

CVector* getWeightforOriginalPolygon( CVector* weightSubd, unsigned int PolySize, unsigned int subdPolySize, const CMatrix& bsplineWeight )
{
	// get the bary-coord for original Poly
	CVector* weight = new CVector( PolySize );

	if ( abs( weightSubd->sum() ) < epsilon )
		return weight;

	for ( unsigned int j = 0; j < PolySize; j++ )
		for ( unsigned int i = 0; i < subdPolySize; i++ )
			(*weight)[j] += bsplineWeight.elem(i, j) * (*weightSubd)[i];

	// normalize weight
	weight->normalize();

	return weight;
}

//'Barycentric' compute the Barycentric coordinate for a point on the boundary or in the interior of an arbitrary polygon
CVector* approBarycentric( QUADTREE::Quadtree& quadTree, const CPoint& X, unsigned int PolySize, const CPoly& subdPoly, const CMatrix& bsplineWeight )
{
	CVector* weightSubd = Barycentric( quadTree, X, subdPoly );

	return getWeightforOriginalPolygon( weightSubd, PolySize, subdPoly.size(), bsplineWeight );

/*
	CVector* weight = new CVector( PolySize );

	if ( abs( weightSubd->sum() ) < epsilon )
		return weight;

	for ( unsigned int j = 0; j < PolySize; j++ )
		for ( unsigned int i = 0; i < subdPoly.size(); i++ )
			(*weight)[j] += bsplineWeight.elem(i, j) * (*weightSubd)[i];

	// normalize weight
	weight->normalize();

	return weight;
*/
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
