#ifndef _BASIC_
#define _BASIC_

#include <string>
#include <fstream>
#include "point.hpp"
#include "vector.hpp"
#include "matrix.hpp"
#include "polygon.hpp"
#include "quadtree.hpp"

using namespace std;


ostream& operator<< ( ostream& output , const CPoint& point );
//ostream& operator<< ( ostream& output , const CPoint3& point );
ostream& operator<< ( ostream& output , const CVector& vector );
ostream& operator<< ( ostream& output , const CMatrix& matrix );
ostream& operator<< ( ostream& output , const CPoly& polygon );

// for CPoint
CPoint operator + ( const CPoint& p1, const CPoint& p2 );
CPoint operator - ( const CPoint& p1, const CPoint& p2 );
CPoint operator * ( double s, const CPoint& p );
CPoint operator / ( const CPoint& p, double s );

// for CVector
CVector operator- ( const CVector& v );
double Dot( const CVector& v1, const CVector& v2 );
CVector operator + ( const CVector& v1, const CVector& v2 );
CVector operator - ( const CVector& v1, const CVector& v2 );
CVector operator * ( double s, const CVector& v );
CVector operator / ( const CVector& v, double s );

// length of a vector
double len( const CVector& v );

// Cross and Det
CVector Cross( const CVector& v );
CPoint Cross( const CPoint& v );
double Det( const CVector& v1, const CVector& v2 );
double Det( const CPoint& v1, const CPoint& v2 );
double Dist( const CPoint& p1, const CPoint& p2, const CPoint& X );


inline double ArcTanh( double x );
inline double Sec( double x );
inline double Csc( double x );


// get angle between v and u
inline double VectorAngle( const CVector& v, const CVector& u );

// get angle between v and CVector(1.0, 0.0)
double getAngle( const CVector& v );


// for CMatrix
CMatrix* operator* ( const CMatrix& M, const CMatrix& N );
vector<CPoint>* operator * ( const CMatrix& M, const vector<CPoint>& controlPoly );


// convert from a polygon to a matrix
CMatrix* Poly2Matrix( const CPoly& Poly );

// convert from a matrix to a polygon
CPoly* Matrix2Poly( const CMatrix& Matrix );

// one B-Spline subdivision
CPoly* sub( const CPoly& Poly );

// multiple times B-Spline subdivision
CMatrix* Bspline( const CMatrix& matrix, unsigned int nSubd );

// get weight for quardratic B-Spline
CMatrix* BsplineWeight( unsigned int m, unsigned int nSubd );


// basic functions
double S( double t, CPoint xx, CPoint p0, CPoint p1, CPoint q0, CPoint q1 );
double Scolinear( double t, CPoint xx, CPoint p0, CPoint p1, CPoint q0, CPoint q1 );
double T( double s, CPoint xx, CPoint p0, CPoint p1, CPoint q0, CPoint q1 );
double W( double t, CPoint xx, CPoint p0, CPoint p1, CPoint q0, CPoint q1 );

CVector IntegrateAngularFptFast( double t1, double t2, const CPoint& p0, const CPoint& p1, const CPoint& q0, const CPoint& q1, const CPoint& X );
CVector IntegrateAngularFptColinearFast( double t1, double t2, const CPoint& p0, const CPoint& p1, const CPoint& q0, const CPoint& q1, const CPoint& X );
CVector IntegrateAngularFptColinearColinearFast( double t1, double t2, const CPoint& p0, const CPoint& p1, const CPoint& q0, const CPoint& q1, const CPoint& X );
CVector IntegrateAngularFptParallelFast( double t1, double t2, const CPoint& p0, const CPoint& p1, const CPoint& q0, const CPoint& q1, const CPoint& X );


// create a set of sample points starting from the left-bottom cornner
vector <CPoint>* CreateSample( const CPoint& LeftBottom, unsigned int w, unsigned int h, double width, double height );

// load a set of sample points from file
vector <CPoint>* LoadSample( const char* filename );

// check if P is on the line segment (A, B)
bool PointInLineSeg( const CPoint& A, const CPoint& B, const CPoint& P );

// check if P is inside the triangle (A, B, C)
bool PointInTriangle( const CPoint& A, const CPoint& B, const CPoint& C, const CPoint& P );

// check if P is inside the rectangle (A, B, C, D)
bool PointInRectangle( const CPoint& A, const CPoint& B, const CPoint& C, const CPoint& D, const CPoint& P );

// 'InsidePolygon' decides if a point is inside an arbitray polygon
bool InsidePolygon( const CPoint& X, const CPoly& Poly );

// 'OnPolygon' decides if a point is on the boundary of an arbitray polygon, if it is, then return the Barycentric coordinate for that boundary point
bool OnPolygon( const CPoint& X, const CPoly& Poly, CVector& weight );

//Get range of t1 and t2 to integrate on current edge (a1, b1)
CVector getRangeOftNew( const CPoint& X, unsigned int a1, unsigned int b1, unsigned int a2, unsigned int b2, const CPoly& Poly );
CVector getRangeOftNewScott( const CPoint& X, unsigned int a1, unsigned int b1, unsigned int a2, unsigned int b2, const CPoly& Poly );

// 'IntegrateAllEdges' will integrate all possible intersected edges for a given input edge jj (a1, b1). This function works on all polygons
void IntegrateAllEdges( QUADTREE::Quadtree& quadTree, unsigned int jj, const CPoint& X, const CPoly& Poly, CVector& weight );

// check if line segment (P0, P1) and (Q0, Q1) are interseting
bool lineIntersect( const CPoint& P0, const CPoint& P1, const CPoint& Q0, const CPoint& Q1 );

#endif