#ifndef _BEZIER_
#define _BEZIER_

#include <utility>
#include "assert.h"
#include "basic.h"
#include "point.hpp"
#include "vector.hpp"
#include "matrix.hpp"
#include "global.hpp"


//------------------------------------------------------------------------
const double curve_distance_epsilon                  = 0.000001;
const double curve_angle_tolerance_epsilon           = 0.01;
const int curve_recursion_limit						 = 3;

CVector operator * ( double s, const CVector& v );

class AdaptiveBezier
{
public:
	AdaptiveBezier( unsigned int _length ) 
	{
		subdPoly = new CMatrix( 0, 2 );
		subdWeight = new CMatrix( 0, _length );
	}

    //------------------------------------------------------------------------
    std::pair < CMatrix*, CMatrix* > recursive_bezier( const CPoint& X1, const CPoint& X2, const CPoint& X3, unsigned level )
    {
        if ( level > curve_recursion_limit )
            return std::make_pair( new CMatrix, new CMatrix );

        CPoint X12   = ( X1 + X2 ) / 2.0;
        CPoint X23   = ( X2 + X3 ) / 2.0;
        CPoint X123  = ( X12 + X23 ) / 2.0;

        double d = fabs( Det( X3 - X1, X3 - X2 ) );

        if ( d < epsilon ) // Collinear case
            return std::make_pair( new CMatrix, new CMatrix );
		else
        {
            // Not a collinear case, but close enough
            if ( d * d <= curve_distance_epsilon * Dot( X3 - X1, X3 - X1 ) )
            {
				// then check the angle change
                //double da = fabs( atan2( X3[1] - X2[1], X3[0] - X2[0] ) - atan2( X2[1] - X1[1], X2[0] - X1[0] ) );
                //if ( da >= PI )
				//	da = 2 * PI - da;
                //if ( da < curve_angle_tolerance_epsilon )
                    return std::make_pair( new CMatrix, new CMatrix );
            }
        }

        std::pair < CMatrix*, CMatrix* > PolyAndWeight1 = recursive_bezier( X1, X12, X123, level + 1 );
        std::pair < CMatrix*, CMatrix* > PolyAndWeight2 = recursive_bezier( X123, X23, X3, level + 1 );

		CMatrix* Poly = new CMatrix;
		CMatrix* Weight = new CMatrix;

		// first part
		if ( PolyAndWeight1.first->nRow() == 0 )
		{
			Poly->addRow( CVector( X1[0], X1[1] ) );
			Poly->addRow( CVector( X12[0], X12[1] ) );
			Weight->addRow( CVector( 1.0, 0.0, 0.0 ) );
			Weight->addRow( CVector( 0.5, 0.5, 0.0 ) );
		}
		else
		{
			Poly->addMatrix( *(PolyAndWeight1.first) );

			CMatrix* tempWeight = *(PolyAndWeight1.second) * quadraticBezierMatrix_New_first;
			Weight->addMatrix( *tempWeight );
		}

		// second part
		if ( PolyAndWeight2.first->nRow() == 0 )
		{
			Poly->addRow( CVector( X123[0], X123[1] ) );
			Poly->addRow( CVector( X23[0], X23[1] ) );
			Weight->addRow( CVector( 0.25, 0.5, 0.25 ) );
			Weight->addRow( CVector( 0.0,  0.5, 0.5 ) );
		}
		else
		{
			Poly->addMatrix( *(PolyAndWeight2.first) );

			CMatrix* tempWeight = *(PolyAndWeight2.second) * quadraticBezierMatrix_New_second;
			Weight->addMatrix( *tempWeight );
		}

		return std::make_pair( Poly, Weight );
    }

    //------------------------------------------------------------------------
    void bezier( const CMatrix& matrix, unsigned int nSubd )
    {
		unsigned int k = matrix.nRow();

		CMatrix* subdFirstPoly = sub_init_2( matrix );
		CMatrix* subdFirstWeight = sub_init_2( CMatrix( k ) );

		// for each segment ( 3 points )
		for ( unsigned int i = 0; i < k; i++ )
		{
			int sta = 2 * i;
			int mid = sta + 1;
			int end = mid + 1;
			if ( end == 2*k )
				end = 0;

			CPoint X1( (*subdFirstPoly)[sta][0], (*subdFirstPoly)[sta][1] );
			CPoint X2( (*subdFirstPoly)[mid][0], (*subdFirstPoly)[mid][1] );
			CPoint X3( (*subdFirstPoly)[end][0], (*subdFirstPoly)[end][1] );

			std::pair < CMatrix*, CMatrix* > PolyAndWeight = recursive_bezier( X1, X2, X3, 0 );

			CMatrix* subdPolyTmp = PolyAndWeight.first;
			CMatrix* subdWeightTmp = PolyAndWeight.second;

			unsigned int M = subdPolyTmp->nRow();

			if ( M == 0 )
			{
				subdPoly->addRow( CVector( X1[0], X1[1] ) );
				subdPoly->addRow( CVector( X2[0], X2[1] ) );

				//unsigned int pi = (k + i - 1) % k;
				unsigned int ni = (i + 1) % k;

				CVector* weight1 = new CVector( k );
				//(*weight1)[ pi ] = 0.25;
				//(*weight1)[ i ] = 0.5;
				//(*weight1)[ ni ] = 0.25;
				(*weight1)[ i ] = 0.5;
				(*weight1)[ ni ] = 0.5;
				subdWeight->addRow( *weight1 );

				CVector* weight2 = new CVector( k );
				//(*weight2)[ i ] = 0.5;
				//(*weight2)[ ni ] = 0.5;
				(*weight2)[ ni ] = 1.0;
				subdWeight->addRow( *weight2 );
			}
			else
			{
				unsigned int N = subdPoly->nRow();
				assert( M == subdWeightTmp->nRow() );

				subdPoly->addMatrix( *subdPolyTmp );

				CMatrix* controlWeight = new CMatrix( 3, k );
				controlWeight->setRow( 0, (*subdFirstWeight)[sta] );
				controlWeight->setRow( 1, (*subdFirstWeight)[mid] );
				controlWeight->setRow( 2, (*subdFirstWeight)[end] );

				subdWeight->addMatrix( *( ( *subdWeightTmp ) * ( *controlWeight ) ) );
			}
		}

		// finally adjust the positions of some vertices
		for ( unsigned int i = 0; i < subdPoly->nRow() / 2; i++ )
		{
			subdPoly->setRow( 2*i+1, 0.5 * (*subdPoly)[2*i+1] + 0.25 * (*subdPoly)[2*i] + 0.25 * (*subdPoly)[ 2*(i+1) % subdPoly->nRow() ] );
			subdWeight->setRow( 2*i+1, 0.5 * (*subdWeight)[2*i+1] + 0.25 * (*subdWeight)[2*i] + 0.25 * (*subdWeight)[ 2*(i+1) % subdPoly->nRow() ] );
		}
	}

	// init sub ( first sub )
	CMatrix* sub_init( const CMatrix& matrix )
	{
		unsigned int k = matrix.nRow();

		CMatrix* subdMatrix = new CMatrix( 2 * k, matrix.nColumn() );

		for ( unsigned int i = 0; i < k; i++ )
		{
			unsigned int pi = (k + i - 1) % k;
			unsigned int ni = (i + 1) % k;

			CMatrix* controlPoly;

			controlPoly = new CMatrix( 3, matrix.nColumn() );
			controlPoly->setRow( 0, matrix[ pi ] );
			controlPoly->setRow( 1, matrix[ i ] );
			controlPoly->setRow( 2, matrix[ ni ] );

			CMatrix* pSubdVerts = quadraticBezierMatrix_New * (*controlPoly);

			subdMatrix->setRow( 2*i, pSubdVerts->getRow(0) );
			subdMatrix->setRow( 2*i+1, pSubdVerts->getRow(1) );
		}

		return subdMatrix;
	}

	// init sub ( first sub )
	CMatrix* sub_init_2( const CMatrix& matrix )
	{
		unsigned int k = matrix.nRow();

		CMatrix* subdMatrix = new CMatrix( 2 * k, matrix.nColumn() );

		for ( unsigned int i = 0; i < k; i++ )
		{
			subdMatrix->setRow( 2*i, 0.5 * matrix[ i ] + 0.5 * matrix[ (i+1) % k ] );
			subdMatrix->setRow( 2*i+1, matrix[ (i+1) % k ] );
		}

		return subdMatrix;
	}

public:
	CMatrix* subdPoly;
	CMatrix* subdWeight;
};

#endif
