#ifndef   _QUADTREE_H_
#define   _QUADTREE_H_

#include <cmath>
#include <vector>
#include "assert.h"

#include "point.hpp"
#include "../array.h"
#include "polygon.hpp"

using namespace std;

CPoint operator + ( const CPoint& p1, const CPoint& p2 );
CPoint operator / ( const CPoint& p, double s );
CVector Cross( const CVector& v );
CPoint Cross( const CPoint& v );
double Dot( const CVector& v1, const CVector& v2 );
CVector operator- ( const CVector& v );
double Dot( const CVector& v1, const CVector& v2 );
double Dot( const CPoint& v1, const CPoint& v2 );
CPoint operator - ( const CPoint& p1, const CPoint& p2 );


namespace QUADTREE
{
  static const int		QUAD_MIN_CAPACITY	= 4;
  static const int		A		= 0;
  static const int		B		= 1;
  static const int		C		= 2;
  static const int		D		= 3;
  static const double	epsilon = 0.00000000001;

  class Quadtree
  {
  public:
	  Quadtree() {_children[0] = NULL;}

    ~Quadtree()
	{
		if ( _children [ 0 ] != NULL )
		{
			for ( unsigned int i = 0; i < 4; i++ )
				delete _children[i];
		}
	}

    /// make a child
    /// A -- B
    /// |    |
    /// D -- C
    Quadtree( const CPoly* _Poly, const CPoint& a, const CPoint& b, const CPoint& c, const CPoint& d )
	{
		Poly = _Poly;
		_children[0] = NULL;

		_setup();
		_corners[A] = a; 
		_corners[B] = b;
		_corners[C] = c; 
		_corners[D] = d;

		_size = abs(a[0] - b[0]);
		_compute_center();
	}

    ///make a root
    Quadtree( const CPoly* _Poly, const vector < unsigned int >& plist )
	{
		//no empty root allowed
		assert( plist.size() );
		_children[0] = NULL;

		Poly = _Poly;

		_setup();

		unsigned int k = Poly->size();

		CPoint min_v = (*Poly)[*(plist.begin())];
		CPoint max_v = (*Poly)[*(plist.begin())];

		for ( unsigned int j = 1; j < plist.size(); j++ )
		{
			unsigned int id = plist[j];

			for ( int i = 0; i < 2; i++ ) // for each coordinates x, y
			{
				if ( (*Poly)[id][i] < min_v[i] )
					min_v[i] = (*Poly)[id][i];
				else if ( (*Poly)[id][i] > max_v[i] )
					max_v[i] = (*Poly)[id][i];

				if ( (*Poly)[(id+1)%k][i] < min_v[i] )
					min_v[i] = (*Poly)[(id+1)%k][i];
				else if ( (*Poly)[(id+1)%k][i] > max_v[i] )
					max_v[i] = (*Poly)[(id+1)%k][i];
			}
		}

		//compute the root corners
		_size = abs( min_v[0] - max_v[0] );

		_corners[A] = CPoint( min_v[0], max_v[1] );
		_corners[B] = CPoint( max_v[0], max_v[1] );
		_corners[C] = CPoint( max_v[0], min_v[1] );
		_corners[D] = CPoint( min_v[0], min_v[1] );

		_compute_center();

		//construct children
		add_polygons( plist );
	}

	// check if pt is in this quad
    inline bool is_in( const CPoint& pt ) const
	{
		return ( pt[0] >= _corners[A][0] && pt[0] <= _corners[B][0] && pt[1] <= _corners[A][1] && pt[1] >= _corners[D][1] );
	}

	// check if edge id is in this quad
    bool is_in( unsigned int id ) const
	{ 
		return is_in( (*Poly)[id] ) || is_in( (*Poly)[(id+1)%Poly->size()] ); 
	}

	// check if edge id is in this quad
    bool is_in_new( unsigned int id ) const
	{ 
		double xmin, xmax, ymin, ymax, projFirst, projTemp;

		unsigned int nid = (id+1)%Poly->size();
 
        if ( (*Poly)[id][0] < (*Poly)[nid][0] )
        {
			xmin = (*Poly)[id][0];
            xmax = (*Poly)[nid][0];
		}
        else
        {
            xmax = (*Poly)[id][0];
            xmin = (*Poly)[nid][0];
        }
 
        if ( xmin > _corners[B][0] ) return false;
        if ( xmax < _corners[D][0] ) return false;
 
        if ( (*Poly)[id][1] < (*Poly)[nid][1] )
        {
            ymin = (*Poly)[id][1];
            ymax = (*Poly)[nid][1];
        }
        else
        {
             ymax = (*Poly)[id][1];
             ymin = (*Poly)[nid][1];
        }
 
        if ( ymin > _corners[B][1] ) return false;
        if ( ymax < _corners[D][1] ) return false;
 
        CPoint norm ( (*Poly)[nid][1] - (*Poly)[id][1], (*Poly)[id][0] - (*Poly)[nid][0] );
        projFirst = norm[0] * ( _corners[0][0] - (*Poly)[id][0] ) + norm[1] * ( _corners[0][1] - (*Poly)[id][1] );
        //CVector norm = -CVector(Cross( (*Poly)[nid] - (*Poly)[id] ));
		//projFirst = Dot( norm, _corners[0] - (*Poly)[id] );

		if ( abs( projFirst ) < epsilon )
			return true;

        for ( int i = 1; i < 4; i++ )
        {
            projTemp = norm[0] * ( _corners[i][0] - (*Poly)[id][0] ) + norm[1] * ( _corners[i][1] - (*Poly)[id][1] );
			//projTemp = Dot( norm, _corners[i] - (*Poly)[id] );

			if ( abs( projTemp ) < epsilon )
				return true;

            if ( projTemp * projFirst < 0 )
				return true;
        }
		return false;
	}


    void add_polygons( const vector < unsigned int >& plist )
	{
		if ( plist.empty() )
			return;

		//if there's only one line in plist or if size <= MIN_SIZE, we add the polygons in our list
		if ( plist.size() < QUAD_MIN_CAPACITY )
		{
			_line_list = plist;
		}
		else
		{
			//create the children
			_create_children();

			//create a subset of plist for each children
			vector< unsigned int > L[4];
			for ( unsigned int j = 0; j < plist.size(); j++ )
			{
				for ( int i = 0; i < 4; i++ )
				{
					if ( _children[i]->is_in_new( plist[j] ) )
						L[i].push_back( plist[j] );
				}
			}

			//add the list to the children
			for ( int i = 0; i < 4; i++ )
				_children[i]->add_polygons( L[i] );
		}
	}

    bool is_leaf(void) const
    { 
//		return _children.size() == 0; 
		return _children[0] == NULL;
	}

    bool is_empty(void) const
	{
		return _line_list.empty();
	}

    const CPoint& get_center(void) const
    { 
		return _center;
	}

	// check if the cone formed by L (with repsect to X) are intersecting quad ( corners )
	bool coneIntersect( const CPoint& X, const CVector& fv, const CVector& sv )
	{
		//     corners[0] --- corners[1]
		//     |					   |
		//     |					   |
		//     |					   |
		//     corners[3] --- corners[2]

//		CPoint boxMin = _corners[D];
//		CPoint boxMax = _corners[B];
		double boxMin [ 2 ], boxMax [ 2 ];
		boxMin [ 0 ] = _corners[D][0];
		boxMin [ 1 ] = _corners[D][1];
		boxMax [ 0 ] = _corners[B][0];
		boxMax [ 1 ] = _corners[B][1];

//		CVector fvn = -Cross( fv );
//		CVector svn = Cross( sv );
		double fvn [ 2 ] = {fv[1], -fv[0]};
		double svn [ 2 ] = {-sv[1], sv[0]};

//		double fmin = Dot( fvn, _corners[0] - X );
		double fmin = fvn[0] * ( _corners[0][0]-X[0]) + fvn[1]*(_corners[0][1]-X[1]);

		for ( int i = 1; i < 4; i++ )
		{
//			double temp = Dot( fvn, _corners[i] - X );
			double temp = fvn[0] * ( _corners[i][0]-X[0]) + fvn[1]*(_corners[i][1]-X[1]);
			if ( temp < fmin )
			{
				fmin = temp;
				//if ( fmin < 0 )
				//	break;
			}
		}
		if ( fmin > 0 )
			return false;

//		double smin = Dot( svn, _corners[0] - X );
		double smin = svn[0] * ( _corners[0][0]-X[0]) + svn[1]*(_corners[0][1]-X[1]);
		for ( int i = 1; i < 4; i++ )
		{
//			double temp = Dot( svn, _corners[i] - X );
			double temp = svn[0] * ( _corners[i][0]-X[0]) + svn[1]*(_corners[i][1]-X[1]);
			if ( temp < smin )
			{
				smin = temp;
				//if ( smin < 0 )
				//	break;
			}
		}
		if ( smin > 0 )
			return false;

		if ( fv[0] * sv[0] > 0 )
		{
			if ( fv[0] < 0 )
			{
				if ( X[0] < boxMin[0] )
					return false;
			}
			else 
			{
				if ( X[0] > boxMax[0] )
					return false;
			}
		}

		if ( fv[1] * sv[1] > 0 )
		{
			if ( fv[1] < 0 )
			{
				if ( X[1] < boxMin[1] )
					return false;
			}
			else 
			{
				if ( X[1] > boxMax[1] )
					return false;
			}
		}

		return true;
	}

	// get visible lines to L with respect to point X
	// remeber to call "std::unique (visible_lines.begin(), visible_lines.end(), compFunction)" afterwards
	void get_visible_lines( const CPoint& X, const CVector& fv, const CVector& sv, Array< unsigned int >& visible_lines )
	{
		if ( !coneIntersect( X, fv, sv ) )
			return;

		if ( is_leaf() )
		{
			for ( unsigned int i = 0; i < _line_list.size(); i++ )
				visible_lines.push_back( _line_list[i] );
		}
		else
		{
			for ( int i = 0; i < 4; i++ )
				_children[i]->get_visible_lines( X, fv, sv, visible_lines );
		}
	}

  private:
	  void _setup(void) { }//_corners.resize(4); }

    void _create_children()
	{
		//     A -- E -- B
		//     |    |    |
		//     F -- G -- H
		//     |    |    |
		//     D -- I -- C
//		_children.resize(4);

		CPoint a( _corners[A] ), b( _corners[B] ), c( _corners[C] ), d( _corners[D] );

		CPoint e = (a + b) / 2.0;
		CPoint f = (a + d) / 2.0;
		CPoint h = (b + c) / 2.0;
		CPoint i = (c + d) / 2.0;
		CPoint g = (f + h) / 2.0;
	    
		_children[A] = new Quadtree( Poly, a, e, g, f );
		_children[B] = new Quadtree( Poly, e, b, h, g );
		_children[C] = new Quadtree( Poly, g, h, c, i );
		_children[D] = new Quadtree( Poly, f, g, i, d );
	}
    
    void _compute_center(void)
    {
		_center = ( _corners[A] + _corners[C] ) / 2.0;
    }

//    vector < CPoint > _corners;
	CPoint _corners [ 4 ];
//	vector < Quadtree* > _children;
	Quadtree *_children [ 4 ];
    vector < unsigned int > _line_list;
	const CPoly *Poly;

    double _size;
    CPoint _center;
  };
  
}

#endif
