// Author: Josiah Manson
// Created: 2004
// Description:
// Uses rdtsc instruction to get the number of cpu cycles elapsed in a section of code.
// Multiple sections can be profiled at the same time. The results are then printed
// to a file.

// Modified: March 28, 2005
// A new method of sorting the entries is used. Split into header and source.

#include "profile.h"

#include <cstdio>
#include <map>
#include <vector>
#include <algorithm>

struct _profileStruct
{
	_profileStruct()
	{
		timesCalled = cyclesTotal = cycleMin = cycleMax = cyclePrev = countStart = 0;
	}

	bool operator<(const _profileStruct &p) const {return false;}

	unsigned __int64 timesCalled;
	unsigned __int64 cyclesTotal;

	unsigned __int64 cycleMin;
	unsigned __int64 cycleMax;
	unsigned __int64 cyclePrev;

	unsigned __int64 countStart;

	string token;
};


map<string, _profileStruct> _profileEntries;
vector<string> _profileTokens;


_declspec(naked) __int64 _fastcall _profileTick()
{
	_asm rdtsc
	_asm ret
}


void profileBegin(string token)
{
	_profileTokens.push_back(token);
	_profileStruct &ent = _profileEntries[token];

	ent.countStart = _profileTick();
}

void profileEnd()
{
	unsigned __int64 countStop = _profileTick();

	string token = _profileTokens[_profileTokens.size() - 1];

	_profileStruct &ent = _profileEntries[token];
	_profileTokens.pop_back();

	// time elapsed
	ent.cyclePrev = countStop - ent.countStart;

	// first time
	if (ent.timesCalled == 0)
	{
		ent.cycleMin = ent.cyclePrev;
		ent.cycleMax = ent.cyclePrev;
		ent.token = token;
	}

	//       counter wrapped                           must have been some glitch (OS task switch?)
	if (ent.cyclePrev > 100000000000000000 || (ent.timesCalled > 3 && ent.cyclePrev > ent.cycleMax * 10))
	{
		ent.timesCalled++;
		ent.cyclesTotal += (ent.cycleMax + ent.cycleMin) / 2;
		return;
	}

	// min max
	if (ent.cyclePrev < ent.cycleMin)
		ent.cycleMin = ent.cyclePrev;

	if (ent.cyclePrev > ent.cycleMax)
		ent.cycleMax = ent.cyclePrev;

	// add up total
	ent.timesCalled++;
	ent.cyclesTotal += ent.cyclePrev;
}

void profilePrint()
{
	// normalize results
	for (int i = 0; i < 1000; i++)
	{
		profileBegin("__norm__");
		profileEnd();
	}

	// print results
	FILE *f = fopen("profile.txt", "a");

	// header
	fprintf(f, "-------total  -----calls  -------max  -------min  -------avg  ----name----\n");
	fprintf(f, "--------------------------------------------------------------------------\n");

	// sort data
	vector<pair<unsigned __int64, _profileStruct> > data;

	for (map<string, _profileStruct>::iterator a = _profileEntries.begin(); a != _profileEntries.end(); a++)
		data.push_back(pair<unsigned __int64, _profileStruct>(a->second.cyclesTotal, a->second));

	sort(data.begin(), data.end());

	// data
	unsigned __int64 min = _profileEntries["__norm__"].cycleMin;

	for (int i = 0; i < data.size(); i++)
	{
		fprintf(f, "%12I64d  ", data[i].second.cyclesTotal - (min * data[i].second.timesCalled)); // total
		fprintf(f, "%10I64d  ", data[i].second.timesCalled); // calls
		fprintf(f, "%10I64d  ", data[i].second.cycleMax - min); // max
		fprintf(f, "%10I64d  ", data[i].second.cycleMin - min); // min
		fprintf(f, "%10I64d  ", (data[i].second.cyclesTotal / data[i].second.timesCalled) - min); // avg
		fprintf(f, data[i].second.token.c_str()); // name
		fprintf(f, "\n");
	}

	// footer
	fprintf(f, "\n\n");

	fclose(f);
}
